import hashlib
import inspect
import json
import logging
import random
import time
import warnings
from typing import Callable

from langchain_core.language_models import BaseChatModel

from ceo.ability.ability import Ability
from ceo.prompt import (
    SchedulerPrompt,
    AnalyserPrompt,
    ExecutorPrompt,
    IntrospectionPrompt,
    RequestResolverPrompt,
    SelfIntroducePrompt
)

log = logging.getLogger('ceo')


class BaseAgent:
    def __init__(self, abilities: list[Callable], brain: BaseChatModel, name: str = '', request: str = ''):
        self._abilities = list()
        self._act_count = 0
        self._name = name
        self._model = brain
        self._request = self._request_by_step = str()
        if request is not None and request != '':
            self._request, self._request_by_step = RequestResolverPrompt(request).invoke(self._model)
        if self._name is None or len(self._name) < 1:
            self._name = self._generate_name()
        self.__prev_results = list()
        self.__schedule = list()
        self._introduction = str()
        self.grant_abilities(abilities)

    @property
    def abilities(self) -> list[Ability]:
        return self._abilities

    @property
    def name(self) -> str:
        return self._name

    @property
    def introduction(self) -> str:
        return self._introduction

    @property
    def brain(self) -> BaseChatModel:
        return self._model

    def __repr__(self):
        return json.dumps(self.to_dict(), ensure_ascii=False)

    def __str__(self):
        return self.__repr__()

    def _generate_name(self) -> str:
        __abilities_ls = [ability.to_dict() for ability in self.abilities]
        __tmp_bytes = f'{__abilities_ls}{time.time()}{random.uniform(0, 10 ** 3)}'.encode('utf-8')
        __tmp_str = hashlib.md5(__tmp_bytes).hexdigest()
        __sample_str = ''.join(random.sample(__tmp_str, 6))
        return f'智能體{__sample_str}型號'

    def to_dict(self) -> dict:
        __model_dict = self._model.dict()
        model_name = __model_dict.get('model_name', __model_dict.get('_type', 'unknown'))
        return {
            "name": self.name,
            "brain": model_name,
            "abilities": [ability.to_dict() for ability in self.abilities]
        }

    def introduce(self, update: bool = False) -> str:
        def get_your_info(*args, **kwargs) -> dict:
            """
            What does this ability do: To get your personal information.
            In any case, the [__SystemAbility__get_your_info] ability can only be used once!!
            :return: All information of yourself.
            """
            _info_dict = self.to_dict()
            if 'brain' in _info_dict:
                del _info_dict['brain']
            return {
                'success': True,
                'info': _info_dict
            }

        get_your_info.__name__ = f'__SystemAbility__{get_your_info.__name__}'
        self.grant_ability(get_your_info)
        if self._introduction == '' or update:
            self._introduction = SelfIntroducePrompt(agent=self).invoke(self._model)
        return self._introduction

    def grant_ability(self, ability: Callable, update_introduction: bool = True):
        for _ability in self.abilities:
            if inspect.getsource(ability) == inspect.getsource(_ability.function):
                return
        self._abilities.append(Ability(ability))
        self.introduce(update_introduction)

    def grant_abilities(self, abilities: list[Callable]):
        for ability in abilities:
            self.grant_ability(ability, update_introduction=False)
        self.introduce(update=True)

    def deprive_ability(self, ability: Callable, update_introduction: bool = True):
        ability = Ability(ability)
        for _ability in self._abilities:
            if _ability.name == ability.name:
                self._abilities.remove(_ability)
        self.introduce(update_introduction)

    def deprive_abilities(self, abilities: list[Callable]):
        for ability in abilities:
            self.deprive_ability(ability, update_introduction=False)
        self.introduce(update=True)

    def plan(self, _log: bool = True) -> list:
        scheduling = SchedulerPrompt(request=self._request_by_step, abilities=self._abilities)
        self.__schedule = scheduling.invoke(self._model)
        if _log:
            log.debug(f'Agent: {self._name}; Schedule: {[_.name for _ in self.__schedule]}; Request: "{self._request}";')
        return self.__schedule

    def reposition(self):
        self.__prev_results = list()
        self.__schedule = list()
        self._act_count = 0
        return self

    def assign(self, request: str):
        self._request, self._request_by_step = (
            RequestResolverPrompt(request=request).invoke(self._model))
        return self.reposition()

    def reassign(self, request: str):
        return self.assign(request)

    def relay(self, request: str, request_by_step: str):
        self._request = request
        self._request_by_step = request_by_step
        return self.reposition()

    def __step_quiet(self) -> str:
        if self._act_count < len(self.__schedule):
            combined_request = {
                'raw_request': self._request,
                'request_by_step': self._request_by_step
            }
            analysing = AnalyserPrompt(
                request=combined_request,
                prev_results=self.__prev_results,
                action=self.__schedule[self._act_count]
            )
            action, args = analysing.invoke(self._model)
            executing = ExecutorPrompt(args=args, action=action)
            action_str = (f'Agent: {self._name}; Action {self._act_count + 1}/{len(self.__schedule)}: '
                          f'{json.dumps(executing.invoke(model=self._model), ensure_ascii=False)};')
            self.__prev_results.append(action_str)
            self._act_count += 1
            log.debug(action_str)
            return action_str
        self.reposition()
        return ''

    def just_do_it(self) -> str | None:
        warnings.warn(
            "This function is deprecated and will be removed in future versions.",
            DeprecationWarning,
            stacklevel=2
        )
        if not self.plan():
            return None
        for act_count in range(len(self.__schedule)):
            self.__step_quiet()
        brief_conclusion, response = IntrospectionPrompt(
            request=self._request,
            history=self.__prev_results,
            self_info=self.introduction
        ).invoke(self._model)
        log.debug(f'Agent: {self._name}; Conclusion: {brief_conclusion};')
        self.reposition()
        return f'{self._name}: {response}'



import copy
import json
import logging
from collections.abc import Iterator

from langchain_core.language_models import BaseChatModel

from ceo.ability.agentic_ability import PREFIX as AGENTIC_ABILITY_PREFIX
from ceo.ability.ability import Ability
from ceo.exception.too_dumb_exception import TooDumbException
from ceo.prompt.prompt import Prompt

log = logging.getLogger('ceo.prompt')


class ExecutorPrompt(Prompt):
    def __init__(self, args: dict, action: Ability, ext_context: str = ''):
        self.action = action
        self.args = copy.deepcopy(args)
        tmp_args = copy.deepcopy(self.args)
        if self.action.name.startswith(AGENTIC_ABILITY_PREFIX):
            del tmp_args['memory']
        prompt = json.dumps({
            "precondition": "Below is an ability shown at <ability> "
                            "and your choice(args) for using the <ability> is shown at <args(choice)>.",
            "task": "Explain what you are going to do.",
            "output_datatype": "text",
            "output_example": "I am trying to open calculator.",
            "ability": self.action.to_dict(),
            "args(choice)": tmp_args
        }, ensure_ascii=False)
        super().__init__(prompt, ext_context)
        log.debug(f'ExecutorPrompt (before): {self.prompt}')

    def explain(self, model: BaseChatModel, stream: bool = False) -> str | Iterator:
        if stream:
            return model.stream(self.prompt)
        resp = model.invoke(self.prompt).content
        log.debug(f'ExecutorResponse (before): {resp}')
        return resp

    def invoke(self, model: BaseChatModel, max_retry: int = 3) -> dict:
        result = self.action.__call__(**self.args)
        tmp_args = copy.deepcopy(self.args)
        if self.action.name.startswith(AGENTIC_ABILITY_PREFIX):
            tmp_args = {'choice': 'Ask for a favor.'}
        prompt = json.dumps({
            "precondition": "Below is an ability shown at <ability>, "
                            "your choice(args) for the <ability> is shown at <args(choice)>, "
                            "result of your using of this <ability> is shown at <result>.",
            "task": "Explain what you have done according to <ability>, <result>, and <args(choice)> "
                    "accurately, comprehensively, and briefly.",
            "ability": self.action.to_dict(),
            "args(choice)": tmp_args,
            "result": str(result),
            "output_format": {
                'ability': '{ability_just_used}',
                'choice': '{choice_just_made}',
                'returns': '{result_just_received}',
                'summarization': '{summarization}'
            },
            "output_example": json.dumps({
                'ability': 'wechat_sender',
                'choice': "{'msg': 'Bonjour'}",
                'returns': 'success',
                'summarization': "I used the wechat_sender ability to wrote a wechat message which says 'Bonjour', "
                                 "the result shows 'success' which indicates success of wechat message sending."
            }, ensure_ascii=False),
            "hint_for_output": 'You must strictly follow the format in <output_format>!! '
                               'You can refer to example in <output_example>!!'
        }, ensure_ascii=False)
        if len(self.ext_context) > 0:
            prompt = Prompt.construct_prompt(prompt, self.ext_context)
        log.debug(f'ExecutorPrompt (after): {prompt}')
        count = 0
        exclamation = '!'
        tmp_prompt = prompt
        keys = ('summarization', 'ability', 'choice', 'returns')
        while True:
            # noinspection DuplicatedCode
            if count > 0:
                if count <= max_retry:
                    log.warning(f'ExecutorAfterPromptWarn: incorrectly formatted. Retry: {count}')
                else:
                    log.warning(f'ExecutorAfterPromptWarn: max retry exceeded.')
                    raise TooDumbException(model)
            count += 1
            res = model.invoke(tmp_prompt).content
            log.debug(f"Executor (after) thought process: \n{res}")
            try:
                correct_format = True
                res_dict: dict = json.loads(res[res.find('{'):res.rfind('}') + 1].strip())
                for _key in keys:
                    if _key not in res_dict.keys():
                        correct_format = False
                if correct_format:
                    break
                tmp_prompt = (f'{prompt}Attention_{count}: '
                              f'You must strictly follow the format in <output_format>{count * 2 * exclamation} '
                              f'You should refer to example in <output_example>{count * 2 * exclamation}')
                tmp_prompt = Prompt.construct_prompt(tmp_prompt, '')
            except json.decoder.JSONDecodeError:
                tmp_prompt = (f'{prompt}Attention_{count}: '
                              f'You must strictly follow the json format in <output_format>{count * 2 * exclamation} '
                              f'You should refer to example in <output_example>{count * 2 * exclamation}')
                tmp_prompt = Prompt.construct_prompt(tmp_prompt, '')
        return res_dict



from typing import Callable


class ClassMethodException(TypeError):
    def __init__(self, func: Callable):
        super().__init__(f'"{func.__qualname__}" is a method. Abilities can only be functions instead of methods')



import asyncio
import copy
import inspect
import json
import threading

from typing_extensions import Callable


class Ability:
    def __init__(self, function: Callable):
        signature = inspect.signature(function)
        doc_str = inspect.getdoc(function)
        self._name: str = function.__name__
        self._description: str | dict = str()
        self._function: Callable = function
        self._parameters: dict = dict()
        self._returns: any = signature.return_annotation
        for name, param in signature.parameters.items():
            self._parameters[name] = str(param.annotation)
        try:
            self._description = json.loads(doc_str)
            self._description = self._description.get('description', self._description)
        except json.decoder.JSONDecodeError:
            self._description = doc_str

    def __repr__(self):
        return json.dumps(self.to_dict(), ensure_ascii=False)

    def __str__(self):
        return self.__repr__()

    def __call__(self, *args, **kwargs):
        if inspect.iscoroutinefunction(self._function):
            __res = None

            def __func(loop: asyncio.AbstractEventLoop):
                nonlocal __res, args, kwargs
                __res = loop.run_until_complete(self._function(*args, **kwargs))

            __thread = threading.Thread(
                target=__func,
                args=(asyncio.new_event_loop(),)
            )
            __thread.start()
            __thread.join(timeout=None)
            return __res
        return self._function(*args, **kwargs)

    def to_dict(self) -> dict:
        param_list: list = list()
        unnecessary_params: tuple = ('args', 'kwargs')
        for name, _ in self._parameters.items():
            if name not in unnecessary_params:
                param_list.append(name)
        return {
            'ability_name': self._name,
            'description': self._description,
            'parameters_required': param_list,
            'returns': str(self._returns)
        }

    @property
    def name(self) -> str:
        return self._name

    @property
    def description(self) -> str:
        return self._description

    @property
    def parameters(self) -> dict:
        return copy.deepcopy(self._parameters)

    @property
    def returns(self) -> any:
        return self._returns

    @property
    def function(self) -> Callable:
        return copy.deepcopy(self._function)



import logging

from .brain.lm import get_openai_model, get_dashscope_model
from .brain.agent import Agent
from .ability import Ability, AgenticAbility
from .util import ability, agentic
from .enum import Personality

__AUTHOR__ = '吴子豪 / Vortez Wohl'
__EMAIL__ = 'vortez.wohl@gmail.com'
__VERSION__ = '0.11.2-preview'

logger = logging.getLogger('ceo')
logger.setLevel(logging.ERROR)
console_handler = logging.StreamHandler()
console_handler.setLevel(logging.DEBUG)
formatter = logging.Formatter('[%(levelname)s] %(asctime)s %(name)s : %(message)s')
console_handler.setFormatter(formatter)
logger.addHandler(console_handler)

logger = logging.getLogger('ceo.prompt')
logger.setLevel(logging.ERROR)
logger = logging.getLogger('ceo.ability')
logger.setLevel(logging.ERROR)
logger = logging.getLogger('ceo.agent')
logger.setLevel(logging.ERROR)



import json
import logging
from collections import OrderedDict

from typing_extensions import override

from ceo.ability import Ability
from ceo.brain.memory_augment import MemoryAugment

PREFIX = '__AgenticAbility__'

from ceo.brain.base_agent import BaseAgent

log = logging.getLogger('ceo.ability')


class AgenticAbility(Ability):
    def __init__(self, agent: BaseAgent | MemoryAugment):
        try:
            _test_mem = agent.memory
        except AttributeError:
            raise TypeError("The 'agent' of AgenticAbility should be instance of 'ceo.Agent'.")
        self._agent = agent
        self.__name__ = f'{PREFIX}talk_to_{agent.name}'
        self.__doc__ = json.dumps({
            "description": {
                "brief_description": f'Initiates a conversation with "{agent.name}" to use its abilities.',
                "detailed_description": f"First, carefully consider and explore {agent.name}'s "
                                        "potential abilities in solving your tasks, "
                                        f"then, if you need {agent.name}'s help, "
                                        "you must tell comprehensively, precisely "
                                        f"and exactly what you need {agent.name} to do.",
                f"self_introduction_from_{agent.name}": agent.introduction,
                "hint": f"By reading <self_introduction_from_{agent.name}>, you can learn what {agent.name} can do, "
                        f"and then decide whether to initiates a conversation with {agent.name} "
                        "according to its abilities.",
                "parameters": [{
                    "request": {
                        "name": "request",
                        "type": "str",
                        "description": f"A brief instruction to be processed by {agent.name}."
                    },
                    "request_by_step": {
                        "name": "request_by_step",
                        "type": "str",
                        "description": "A comprehensive, precise and exact instruction "
                                       f"to be processed by {agent.name}."
                    },
                    "memory": {
                        "name": "memory",
                        "type": "dict",
                        "description": "/"
                    }
                }],
                "returns": {
                    "type": "str",
                    "description": f"{agent.name}'s response to your instruction."
                }
            }
        }, ensure_ascii=False)
        super().__init__(self)
        log.debug(f'Agent dispatcher generated. {self.__name__}: {self.__doc__}')

    @override
    def __call__(self, request: str, request_by_step: str, memory: OrderedDict, *args, **kwargs) -> str:
        self._agent.relay(request_by_step=request_by_step, request=request)
        self._agent.bring_in_memory(memory)
        result = self._agent.just_do_it()
        if isinstance(result, dict):
            if 'conclusion' in result.keys():
                del result['conclusion']
            if 'misc' in result.keys():
                del result['misc']
            return json.dumps(result, ensure_ascii=False)
        return result
